-- Information model for a unit, abstracts the event handling parts of
-- the unit frame into its own object.
--
-- Model values
--
--   shown -- Boolean (true/false) - should ring be visible.
--
--   The following are only updated when shown is true
--   
--   health / healthMax / healthPct - Health value, max, percent (0-1)
--   mana / manaMax / manaPct       - Mana value, max, percent (0-1)
--   powerType                      - Power type number
--   isRage                         - Power type is rage (Do i need this?)
--   unitName                       - Unit's name

local MODEL = {};

local function NOTIFY(self)
   local watcher = self.watcher;
   if (watcher) then
      watcher:NotifyInfoUpdate(self);
   end
end

-- Check and update HEALTH
-- Return hasChanged
local function CHECK_Health(self, unit)
   local h = UnitHealth(unit);
   local mh = UnitHealthMax(unit);
   if (UnitIsDeadOrGhost(unit)) then
      h = 0;
   end
   if (h == self.health and mh == self.healthMax) then
      return;
   end
   self.health = h;
   self.healthMax = mh;
   if (mh < 1) then
      mh = 1;
   end
   h = h / mh;
   if (h < 0) then
      h = 0;
   elseif (h > 1) then
      h = 1;
   end
   self.healthPct = h;
   return true;
end

-- Check and update MANA
-- Return hasChanged
local function CHECK_Mana(self, unit)
   local m = UnitMana(unit);
   local mm = UnitManaMax(unit);
   if (m == self.mana and mm == self.manaMax) then
      return;
   end
   self.mana = m;
   self.manaMax = mm;
   if (mm < 1) then
      mm = 1;
   end
   m = m / mm;
   if (m < 0) then
      m = 0;
   elseif (m > 1) then
      m = 1;
   end
   self.manaPct = m;
   return true;
end

-- Check and update POWER TYPE
-- Return hasChanged
local function CHECK_PowerType(self, unit)
   local pType = UnitPowerType(unit) or 0;
   if (self.pType == pType) then
      return nil;
   end
   
   self.powerType = pType;
   self.isRage = ((pType == 1) or nil);
   return true;
end

-- Check and update NAME
-- Return hasChanged
local function CHECK_Name(self, unit)
   local name = UnitName(unit) or "";
   if (name == self.unitName) then
      return nil;
   end
   
   self.unitName = name;
   return true;
end


-- Check to see if this ring should be shown (does unit exist)
-- Return hasChanged, isShown
local function CHECK_Shown(self, unit)
   local exists = UnitExists(unit);
   local shown = self.shown;
   if ((exists and shown) or not (exists or shown)) then
      return nil, shown;
   end
   self.shown = (exists and true) or false;
   return true, exists;
end

-- Check shown state, and if visible, everything else. Call
-- Notify function if anything changed.
--
-- If you're adding any cascading checks, remember to always put the
-- function call BEFORE the or, otherwise lua wont evaluate it if the
-- changed flag is true. 
local function CHECK_All(self, unit, noNotify)
   local change,shown = CHECK_Shown(self, unit);
   if (shown) then
      change = CHECK_PowerType(self, unit) or change;
      change = CHECK_Health(self, unit) or change;
      change = CHECK_Mana(self, unit) or change;
      change = CHECK_Name(self, unit) or change;
   end
   if (change and (not noNotify)) then NOTIFY(self) end
end

-- Event handler, arguments are hopefully quite obvious
-- 
-- If you're adding any cascading checks, remember to always put the
-- function call BEFORE the or, otherwise lua wont evaluate it if the
-- changed flag is true. Also remember to update CHECK_All
function MODEL:OnEvent(event, arg1)
   local unit = self.unit;

   if (event == "PLAYER_ENTERING_WORLD") then
      CHECK_All(self, unit);
      return;
   end

   if (event == "UNIT_PET") then
      if (arg1 == self.owner) then
	 CHECK_All(self, unit);
      end
      return;
   end

   if (event == "PLAYER_TARGET_CHANGED") then
      -- No need to check the unit, since this event is only registered
      -- for the target.
      CHECK_All(self, unit);
      return;
   end

   if (event == "PARTY_MEMBERS_CHANGED") then
      -- No need to check the unit, since this event is only registered
      -- for party units.
      local name = UnitName(unit);
      if (name) then
	 if ((name == self.unitName) and (self.shown) and 
	     (name ~= UNKNOWNOBJECT)) then
	    return;
	 end
      else
	 if (not self.shown) then
	    return;
	 end
      end
      CHECK_All(self, unit);
      return;
   end

   -- All of the events below this point are for THIS unit, they also
   -- all require the unit to exist
   if ((arg1 ~= unit) or (not self.shown)) then
      return;
   end

   if (event == "UNIT_NAME_UPDATE") then
      if (CHECK_Name(self, unit)) then NOTIFY(self) end;
      return;
   end

   if (event == "UNIT_DISPLAYPOWER") then
      local changed = CHECK_PowerType(self, unit);
      changed = CHECK_Mana(self, unit) or changed;
      if (changed) then NOTIFY(self) end;
      return;
   end

   if (event == "UNIT_AURA") then
      -- TODO - Avoid repeated updates
      local changed = CHECK_PowerType(self, unit);
      changed = CHECK_Mana(self, unit) or changed;
      changed = CHECK_Health(self, unit) or changed;
      if (changed) then NOTIFY(self) end;
      return;
   end

   if ((event == "UNIT_HEALTH") or (event == "UNIT_MAXHEALTH")) then
      if (CHECK_Health(self, unit)) then NOTIFY(self) end
      return;
   end

   -- Not sure why mana's down here on its own, but it is.
   if (CHECK_Mana(self, unit)) then NOTIFY(self) end
end

-- Initialize a model for the specified unit, with the given frame as
-- the event source
function MODEL:Initialize(frame, unit)
   local targetUnit = (unit == "target");
   local partyUnit  = (string.sub(unit, 1, 5) == "party");
   local petUnit    = (unit=='pet' or string.sub(unit, -4, -1) == "pet");

   self.unit = unit;
   self.frame = frame;

   if (petUnit) then
      if (unit == "pet") then
	 self.owner = "player";
      else
	 self.owner = string.gsub(unit, "pet", "");
      end
   end
   
   frame:RegisterEvent("PLAYER_ENTERING_WORLD");
   frame:RegisterEvent("UNIT_AURA");
   if (petUnit) then
      frame:RegisterEvent("UNIT_PET");
   end
   if (targetUnit) then
      frame:RegisterEvent("PLAYER_TARGET_CHANGED");
   end
   if (partyUnit) then
      frame:RegisterEvent("PARTY_MEMBERS_CHANGED");
   end
   frame:RegisterEvent("UNIT_NAME_UPDATE");
   frame:RegisterEvent("UNIT_MANA");
   frame:RegisterEvent("UNIT_MAXMANA");
   frame:RegisterEvent("UNIT_FOCUS");
   frame:RegisterEvent("UNIT_MAXFOCUS");
   frame:RegisterEvent("UNIT_ENERGY");
   frame:RegisterEvent("UNIT_MAXENERGY");
   frame:RegisterEvent("UNIT_RAGE");
   frame:RegisterEvent("UNIT_MAXRAGE");
   frame:RegisterEvent("UNIT_HEALTH");
   frame:RegisterEvent("UNIT_MAXHEALTH");
   frame:RegisterEvent("UNIT_DISPLAYPOWER");
   
   CHECK_All(self, unit, true);
end

function MODEL:CreateOnEvent(noApply)
   local func = function() self:OnEvent(event, arg1) end;
   if (not noApply) then
      local frame = self.frame;
      frame:SetScript("OnEvent", func);
   end
   return func;
end

function StatRingsUnitModel_New(watcher)
   local new = {};
   for k,v in pairs(MODEL) do 
      new[k] = v;
   end
   new.watcher = watcher;
   return new;
end
